import functools

import venusian


class tween_config(object):
    """ A decorator which allows developers to annotate tween factories. """

    venusian = venusian

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.info = None

    def configure(self, tween_factory, context, name, obj):
        # Reset info just in case something funky happens.
        info = self.info
        self.info = None

        module_name = tween_factory.__module__
        callable_name = tween_factory.__name__
        factory_string = module_name + '.' + callable_name

        context.config.with_package(info.module).add_tween(
            factory_string,
            *self.args,
            **self.kwargs
        )

    def __call__(self, tween_factory):
        self.info = self.venusian.attach(
            tween_factory,
            functools.partial(self.configure, tween_factory),
            category='opbeat_pyramid',
        )

        return tween_factory



import collections
import functools
import mock
import os
import unittest

from pyramid import httpexceptions
from pyramid import testing

from opbeat_pyramid import subscribers


MOCK_APP_ID = 'mock app id'
MOCK_SECRET_TOKEN = 'mock secret token'

SETTING_NAME = 'mock_setting'
EXPECTED_VALUE = 'Expected Value'
UNEXPECTED_VALUE = 'Unexpected Value'

ENV_SETTING_NAME = 'mock_env_setting'
EXPECTED_ENV_VALUE = 'Expected Env Value'
UNEXPECTED_ENV_VALUE = 'Unexpected Env Value'

DEFAULT_VALUE = 'Default Value'

MOCK_QUERY_STRING = 'mock&query=string'

MOCK_USER_AGENT = (
    'Mozilla/5.0 (iPhone; CPU iPhone OS 5_0 like Mac OS X) '
    'AppleWebKit/534.46 (KHTML, like Gecko) '
    'Version/5.1 Mobile/9A334 Safari/7534.48.3'
)

MockRequestEvent = collections.namedtuple('RequestEvent', 'request')


class OpbeatSubscribersTestCase(unittest.TestCase):
    def setUp(self):
        setting_key = 'opbeat.' + SETTING_NAME

        self.config = testing.setUp()
        self.request = testing.DummyRequest(self.config)

        self.request.client_addr = '0.0.0.0'
        self.request.exc_info = None
        self.request.host = 'example.com:443'
        self.request.scheme = 'https'
        self.request.user_agent = MOCK_USER_AGENT

        self.request.matched_route = mock.MagicMock()
        self.request.matched_route.name = 'example_view'

        self.settings = self.request.registry.settings = {
            setting_key: EXPECTED_VALUE,
            'mock_setting': UNEXPECTED_VALUE,
            'opbeat.enabled': 'true',
            'opbeat.module_name': 'mock',
            'opbeat.app_id': 'mock app id',
            'opbeat.secret_token': MOCK_SECRET_TOKEN,
            'opbeat.organization_id': MOCK_SECRET_TOKEN,
        }

        os.environ['OPBEAT_MOCK_ENV_SETTING'] = EXPECTED_ENV_VALUE
        os.environ['MOCK_ENV_SETTING'] = UNEXPECTED_ENV_VALUE

    def tearDown(self):
        del os.environ['OPBEAT_MOCK_ENV_SETTING']
        del os.environ['MOCK_ENV_SETTING']

    def test_get_opbeat_setting_gets_value_from_request_settings(self):
        value = subscribers.get_opbeat_setting(self.request, SETTING_NAME)
        self.assertIs(value, EXPECTED_VALUE)

    def test_get_opbeat_setting_gets_value_from_env_settings(self):
        value = subscribers.get_opbeat_setting(self.request, ENV_SETTING_NAME)
        self.assertEqual(value, EXPECTED_ENV_VALUE)

    def test_get_opbeat_setting_prefers_environment_over_settings(self):
        local_name = 'opbeat_some_setting'
        local_name_upper = local_name.upper()

        os.environ[local_name_upper] = EXPECTED_ENV_VALUE
        value = subscribers.get_opbeat_setting(self.request, 'some_setting')
        del os.environ[local_name_upper]

        self.assertEqual(value, EXPECTED_ENV_VALUE)

    def test_get_opbeat_setting_returns_default_when_not_set(self):
        self.assertIs(DEFAULT_VALUE, subscribers.get_opbeat_setting(
            self.request,
            'unknown_setting',
            default=DEFAULT_VALUE,
        ))

    def test_get_opbeat_setting_raises_ValueError_without_a_default(self):
        break_shit = functools.partial(
            subscribers.get_opbeat_setting,
            self.request,
            'unknown_setting',
        )

        self.assertRaises(ValueError, break_shit)

    @mock.patch('opbeat.Client')
    def test_opbeat_client_factory_returns_an_opbeat_client(self, Client):
        MOCK_RETURN_VALUE = {}
        Client.return_value = MOCK_RETURN_VALUE

        result = subscribers.opbeat_client_factory(self.request)
        Client.assert_called_once()

        self.assertIs(result, MOCK_RETURN_VALUE)

    @mock.patch('opbeat.Client')
    def test_opbeat_client_factory_caches_by_app_id(self, Client):
        subscribers.opbeat_client_factory(self.request)
        subscribers.opbeat_client_factory(self.request)
        Client.assert_called_once()

    @mock.patch('opbeat.Client')
    def test_opbeat_client_factory_wont_cache_separate_apps(self, Client):
        subscribers.opbeat_client_factory(self.request)

        self.settings['opbeat.app_id'] = 'Another App ID'
        subscribers.opbeat_client_factory(self.request)

        self.assertEqual(Client.call_count, 2)

    def test_setting_is_enabled_returns_true_for_truthy_values(self):
        is_enabled = functools.partial(
            subscribers.setting_is_enabled,
            self.request,
            'truthy_value',
        )

        self.settings['opbeat.truthy_value'] = True
        self.assertTrue(is_enabled())

        self.settings['opbeat.truthy_value'] = 'true'
        self.assertTrue(is_enabled())

        self.settings['opbeat.truthy_value'] = 'on'
        self.assertTrue(is_enabled())

        self.settings['opbeat.truthy_value'] = 'yes'
        self.assertTrue(is_enabled())

    def test_setting_is_enabled_returns_false_for_falsy_values(self):
        is_enabled = functools.partial(
            subscribers.setting_is_enabled,
            self.request,
            'truthy_value',
        )

        self.settings['opbeat.truthy_value'] = False
        self.assertFalse(is_enabled())

        self.settings['opbeat.truthy_value'] = 'false'
        self.assertFalse(is_enabled())

        self.settings['opbeat.truthy_value'] = 'off'
        self.assertFalse(is_enabled())

        self.settings['opbeat.truthy_value'] = 'no'
        self.assertFalse(is_enabled())

    def test_get_request_module_name_returns_module_name_from_settings(self):
        module_name = subscribers.get_request_module_name(self.request)
        self.assertEqual(module_name, 'mock')

    def test_get_request_module_name_returns_default_if_setting_missing(self):
        del self.settings['opbeat.module_name']

        module_name = subscribers.get_request_module_name(self.request)
        self.assertEqual(module_name, 'UNKNOWN_MODULE')

    def test_get_safe_settings_returns_settings_without_unsafe_keywords(self):
        MOCK_KEYS = [
            'unsafe_token', 'SECRET_ID', 'MockPassword',
            'passphrase', 'private_item', 'local_key',
        ]

        del self.settings['opbeat.secret_token']

        for key in MOCK_KEYS:
            absolute_key = 'opbeat.' + key
            self.settings[absolute_key] = 'mock unsafe token'

        results = subscribers.get_safe_settings(self.request)

        num_results = len(results.keys())
        num_settings = len(self.settings)
        num_bad_keys = len(MOCK_KEYS)

        for key in MOCK_KEYS:
            self.assertNotIn('opbeat.' + key, results)

        self.assertEqual(num_results, num_settings - num_bad_keys)

    def test_should_ignore_HttpException_returns_false_by_default(self):
        mock_exc_info = [None, httpexceptions.HTTPNotFound()]

        self.assertFalse(subscribers.should_ignore_exception(
            self.request,
            mock_exc_info,
        ))

    def test_should_ignore_HttpException_returns_true_when_enabled(self):
        settings = self.settings
        settings['opbeat.ignore_http_exceptions'] = 'true'
        mock_exc_info = [None, httpexceptions.HTTPNotFound()]

        self.assertTrue(subscribers.should_ignore_exception(
            self.request,
            mock_exc_info,
        ))

    def test_should_not_ignore_exceptions_unless_they_are_HttpExceptions(self):
        self.settings['opbeat.ignore_http_exceptions'] = 'true'
        mock_exc_info = [None, ValueError()]

        self.assertFalse(subscribers.should_ignore_exception(
            self.request,
            mock_exc_info,
        ))

    def test_should_ignore_HttpException_returns_false_when_disabled(self):
        self.settings['opbeat.ignore_http_exceptions'] = 'false'
        mock_exc_info = [None, httpexceptions.HTTPNotFound()]

        self.assertFalse(subscribers.should_ignore_exception(
            self.request,
            mock_exc_info,
        ))

    @mock.patch('opbeat.Client')
    def test_capture_exception_ignores_errors_from_opbeat_client(self, Client):
        # NOTE: This uses a catchall which seems bad, but we do it right now.

        client = mock.MagicMock()
        client.capture_exception.side_effect = ValueError()
        Client.return_value = client

        mock_exc_info = [None, ValueError()]

        self.assertRaises(ValueError, client.capture_exception)
        subscribers.capture_exception(self.request, mock_exc_info, extra={})
        self.assertEqual(2, client.capture_exception.call_count)

    @mock.patch('opbeat.Client')
    def test_handle_exception_sends_an_exception_to_opbeat(self, Client):
        client = mock.MagicMock()
        Client.return_value = client

        self.request.query_string = MOCK_QUERY_STRING
        mock_exc_info = [None, ValueError()]

        subscribers.handle_exception(self.request, mock_exc_info)

        expected_data = {
            'http': {
                'url': 'https://example.com:443/',
                'method': 'GET',
                'query_string': MOCK_QUERY_STRING,
            }
        }

        mock_extra_metadata = {
            'client_ip_address': self.request.client_addr,
            'logging_successful': 'true',
            'url': self.request.url,
            'user_agent': self.request.user_agent,
        }

        mock_extra_metadata.update(subscribers.get_safe_settings(self.request))

        client.capture_exception.assert_called_once_with(
            mock_exc_info,
            data=expected_data,
            extra=mock_extra_metadata,
        )

    @mock.patch('opbeat.Client')
    @mock.patch('opbeat_pyramid.subscribers.should_ignore_exception')
    def test_handle_exception_ignores_ignored_exceptions(self, mock, Client):
        mock.return_value = True
        mock_exc_info = [None, httpexceptions.HTTPNotFound()]

        client = mock.MagicMock()
        Client.return_value = client

        subscribers.handle_exception(self.request, mock_exc_info)
        mock.assert_called_once_with(self.request, mock_exc_info)
        client.capture_exception.assert_not_called()

    def test_get_exception_for_request_returns_exc_info_if_not_None(self):
        self.request.exc_info = ''
        exc_info = subscribers.get_exception_for_request(self.request)
        self.assertIs(exc_info, self.request.exc_info)

    @mock.patch('sys.exc_info')
    def test_get_exception_for_request_uses_sys_as_fallback(self, _exc_info):
        self.request.exc_info = None
        _exc_info.return_value = {}
        exc_info = subscribers.get_exception_for_request(self.request)
        self.assertIs(exc_info, _exc_info.return_value)

    def test_opbeat_tween_gets_response_if_no_error_occured(self):
        mock_response = {}

        handler = mock.MagicMock()
        handler.return_value = mock_response

        response = subscribers.opbeat_tween(
            handler,
            self.request.registry,
            self.request,
        )

        self.assertIs(response, mock_response)
        handler.assert_called_once_with(self.request)

    @mock.patch('sys.exc_info')
    @mock.patch('opbeat_pyramid.subscribers.handle_exception')
    def test_opbeat_tween_raises_handler_exceptions(self, handle_exc, mock):
        exc_info = mock.return_value = [None, ValueError()]

        handler = mock.MagicMock()
        handler.side_effect = exc_info[1]

        break_shit = functools.partial(
            subscribers.opbeat_tween,
            handler,
            self.request.registry,
            self.request,
        )

        self.assertRaises(ValueError, break_shit)
        handle_exc.assert_called_once_with(self.request, exc_info)

    def test_get_status_code_returns_status_code_from_response(self):
        code = 415
        self.request.response.status_code = code
        self.assertEqual(code, subscribers.get_status_code(self.request))

    def test_get_status_code_returns_exception_code_if_not_None(self):
        self.request.exc_info = [None, httpexceptions.HTTPNotFound()]
        self.assertEqual(404, subscribers.get_status_code(self.request))

    def test_get_route_name_uses_view_name_if_available(self):
        self.request.view_name = 'example.view'

        # Ensure these are set to ensure view_name is preferred
        self.request.matched_route.name = 'something_bad_happened'

        route_name = subscribers.get_route_name(self.request)
        self.assertIs(self.request.view_name, route_name)

    def test_get_route_name_uses_matched_route_if_available(self):
        route_name = subscribers.get_route_name(self.request)
        self.assertEqual('mock.example_view', route_name)

    def test_get_route_name_uses_unknown_route_name_setting_as_fallback(self):
        MOCK_UNKNOWN_ROUTE_NAME = 'Mock Unknown Route'

        self.request.view_name = None
        self.request.matched_route = None

        self.settings['opbeat.unknown_route_name'] = MOCK_UNKNOWN_ROUTE_NAME
        route_name = subscribers.get_route_name(self.request)
        self.assertEqual(route_name, MOCK_UNKNOWN_ROUTE_NAME)

    @mock.patch('opbeat_pyramid.subscribers.opbeat_tween')
    def test_opbeat_tween_factory_returns_a_curried_tween_function(self, mock):
        handler = {}

        registry = self.request.registry

        result = subscribers.opbeat_tween_factory(handler, registry)

        mock.assert_not_called()
        result(self.request)
        mock.assert_called_once_with(handler, registry, self.request)

    @mock.patch('opbeat.Client')
    def test_on_request_begin_starts_a_transaction(self, Client):
        client = mock.MagicMock()
        Client.return_value = client

        self.request.add_finished_callback = mock.MagicMock()

        mock_event = MockRequestEvent(self.request)
        subscribers.on_request_begin(mock_event)

        client.begin_transaction.assert_called_once()

        self.request.add_finished_callback.assert_called_once_with(
            subscribers.on_request_finished,
        )

    @mock.patch('opbeat.Client')
    def test_on_request_begin_is_a_noop_if_opbeat_disabled(self, Client):
        client = mock.MagicMock()
        Client.return_value = client

        self.settings['opbeat.enabled'] = False
        self.request.add_finished_callback = mock.MagicMock()

        mock_event = MockRequestEvent(self.request)
        subscribers.on_request_begin(mock_event)

        client.begin_transaction.assert_not_called()
        self.request.add_finished_callback.assert_not_called()

    @mock.patch('opbeat.Client')
    def test_on_request_finished_ends_the_current_transaction(self, Client):
        client = mock.MagicMock()
        Client.return_value = client

        self.request._opbeat_client = client

        client.end_transaction.assert_not_called()
        subscribers.on_request_finished(self.request)

        client.end_transaction.assert_called_once_with(
            'mock.example_view',
            200,
        )

    @mock.patch('opbeat.Client')
    def test_on_request_finished_does_nothing_if_disabled(self, Client):
        client = mock.MagicMock()
        Client.return_value = client

        client.end_transaction.assert_not_called()
        subscribers.on_request_finished(self.request)

        client.end_transaction.assert_not_called()



import mock
import unittest

from opbeat_pyramid import tweens


class OpbeatSubscribersTestCase(unittest.TestCase):
    @mock.patch('venusian.attach')
    def test_tween_returns_original_wrapped_object(self, attach):
        handler = mock.MagicMock()
        wrapped = tweens.tween_config()(handler)

        handler.assert_not_called()
        self.assertIs(wrapped, handler)

    @mock.patch('venusian.attach')
    @mock.patch('opbeat_pyramid.tweens.tween_config.configure')
    def test_tween_attaches_configure_to_venusian(self, configure, attach):
        handler = mock.MagicMock()
        tweens.tween_config()(handler)

        attach.assert_called_once()

        call_args = attach.call_args[0]
        self.assertIs(handler, call_args[0])

        configure.assert_not_called()
        call_args[1]()
        configure.assert_called_once_with(handler)

    def test_configure_actually_configures_a_tween(self):
        MOCK_MODULE = 'mock_module'

        mock_args = [1, 2, 3]
        mock_kwargs = {'mock': 'kwarg'}

        config = tweens.tween_config(*mock_args, **mock_kwargs)
        config.info = mock.MagicMock()  # Mock out venusian info
        config.info.module = MOCK_MODULE

        context = mock.MagicMock()
        config_instance = mock.MagicMock()
        context.config.with_package.return_value = config_instance

        factory = mock.MagicMock()
        obj = mock.MagicMock()

        factory.__module__ = MOCK_MODULE
        factory.__name__ = 'factory'

        config.configure(factory, context, 'example', obj)

        config_instance.add_tween.assert_called_once_with(
            'mock_module.factory',
            *mock_args,
            **mock_kwargs
        )



__VERSION__ = '1.0.11'


def _should_ignore_module(module_name):
    return module_name.endswith('_spec')


def includeme(config, module_name='opbeat_pyramid'):
    """ Extensibility function for using this module with any Pyramid app. """

    config.scan(module_name, ignore=_should_ignore_module)



import mock
import unittest

import opbeat_pyramid


class OpbeatSubscribersTestCase(unittest.TestCase):
    def test_should_ignore_module_does_not_ignore_by_default(self):
        self.assertFalse(opbeat_pyramid._should_ignore_module(
            'opbeat_pyramid.subscribers',
        ))

    def test_includeme_scans_with_the_expected_arguments(self):
        config = mock.MagicMock()
        opbeat_pyramid.includeme(config)

        config.scan.assert_called_once()
        config.scan.assert_called_with(
            opbeat_pyramid.__name__,
            ignore=opbeat_pyramid._should_ignore_module,
        )



import logging
import functools
import opbeat
import os
import sys


import pyramid.tweens

from opbeat.instrumentation import control

from pyramid import events
from pyramid import httpexceptions
from pyramid import settings

from opbeat_pyramid import tweens


control.instrument()


DEFAULT_UNKNOWN_ROUTE_TEXT = 'Unknown Route'
NO_DEFAULT_PROVIDED = {}
OPBEAT_SETTING_PREFIX = 'opbeat.'
TRUTHY_VALUES = {True, 'true', 'yes', 'on'}


DEFAULT_UNSAFE_SETTINGS_PHRASES = (
    'token,password,passphrase,'
    'secret,private,key'
)


logger = logging.getLogger(__name__)


def get_opbeat_setting(request, name, default=NO_DEFAULT_PROVIDED):
    setting_name = OPBEAT_SETTING_PREFIX + name

    env_setting_name = setting_name.replace('.', '_').upper()
    environment_override = os.environ.get(env_setting_name)

    if environment_override:
        return environment_override

    result = request.registry.settings.get(setting_name, default)

    if result is NO_DEFAULT_PROVIDED:
        raise ValueError('Setting ' + setting_name + ' is required.')

    return result


def get_opbeat_client_cache(request):
    if not hasattr(request.registry, '_opbeat_clients'):
        request.registry._opbeat_clients = {}

    return request.registry._opbeat_clients


def create_opbeat_client(request, app_id):
    secret_token = get_opbeat_setting(request, 'secret_token')
    organization_id = get_opbeat_setting(request, 'organization_id')

    return opbeat.Client(
        secret_token=secret_token,
        organization_id=organization_id,
        app_id=app_id,
    )


def opbeat_client_factory(request):
    clients = get_opbeat_client_cache(request)

    app_id = get_opbeat_setting(request, 'app_id')
    client = clients.get(app_id)

    if client:
        return client

    clients[app_id] = create_opbeat_client(request, app_id)

    return clients[app_id]


def setting_is_enabled(request, setting_name):
    return settings.asbool(get_opbeat_setting(request, setting_name, False))


def is_opbeat_enabled(request):
    return settings.asbool(get_opbeat_setting(
        request,
        'enabled',
        default=False,
    ))


def get_request_module_name(request):
    return get_opbeat_setting(request, 'module_name', default='UNKNOWN_MODULE')


def get_unsafe_settings_phrases(request):
    unsafe_phrases = get_opbeat_setting(
        request,
        'unsafe_setting_phrases',
        default=DEFAULT_UNSAFE_SETTINGS_PHRASES,
    )

    return set(unsafe_phrases.split(','))


def is_unsafe_phrase(phrase, unsafe_phrases):
    for key in unsafe_phrases:
        if key.lower() in phrase.lower():
            return True

    return False


def get_safe_settings(request):
    result = {}

    for key in request.registry.settings.keys():
        if not is_unsafe_phrase(key, get_unsafe_settings_phrases(request)):
            result[key] = request.registry.settings[key]

    return result


def should_ignore_exception(request, exc):
    if not is_http_exception(exc):
        return False

    return settings.asbool(get_opbeat_setting(
        request,
        'ignore_http_exceptions',
        default=False,
    ))


def capture_exception(request, exc_info, extra):
    client = opbeat_client_factory(request)

    data = {
        'http': {
            'url': get_full_request_url(request),
            'method': request.method,
            'query_string': request.query_string,
        }
    }

    try:
        return client.capture_exception(exc_info, data=data, extra=extra)

    except Exception:
        # NOTE: This should not be allowed until we know which exception we are
        # looking for here.
        pass


def get_full_request_url(request):
    """ Get the full URL for a given request. """

    return request.scheme + '://' + request.host + request.path


def handle_exception(request, exc_info):
    if should_ignore_exception(request, exc_info):
        return

    details = get_safe_settings(request)

    details.update({
        'client_ip_address': request.client_addr,
        'logging_successful': 'true',
        'url': request.url,
        'user_agent': request.user_agent,
    })

    logger.error('An error occured. Sending to opbeat.', exc_info=exc_info)
    return capture_exception(request, exc_info, details)


def get_exception_for_request(request):
    exc_info = getattr(request, 'exc_info', None)

    if exc_info is not None:
        return exc_info

    return sys.exc_info()


def opbeat_tween(handler, registry, request):
    try:
        response = handler(request)
    except Exception:
        handle_exception(request, sys.exc_info())
        raise

    exc_info = get_exception_for_request(request)
    if exc_info is not None:
        handle_exception(request, exc_info)

    return response


@tweens.tween_config(over=[
    pyramid.tweens.EXCVIEW,
    'pyramid_tm.tm_tween_factory',
])
def opbeat_tween_factory(handler, registry):
    return functools.partial(opbeat_tween, handler, registry)


def is_http_exception(exc_info):
    if exc_info and exc_info[1]:
        exc_info = exc_info[1]

    return isinstance(exc_info, httpexceptions.HTTPException)


def get_status_code(request):
    # Handles an edge-case where `request.response` isn't the actual response.
    # This can occur whenever a view needs to create a new response instead of
    # using the one attached to the request.
    if is_http_exception(request.exc_info):
        return request.exc_info[1].code

    return request.response.status_code


def get_route_name(request):
    if request.view_name:
        return request.view_name

    elif request.matched_route and request.matched_route.name:
        module_name = get_request_module_name(request)
        return module_name + '.' + request.matched_route.name

    return get_opbeat_setting(
        request,
        'unknown_route_name',
        DEFAULT_UNKNOWN_ROUTE_TEXT,
    )


@events.subscriber(events.NewRequest)
def on_request_begin(event):
    request = event.request

    if not is_opbeat_enabled(request):
        return

    client = request._opbeat_client = opbeat_client_factory(request)
    client.begin_transaction(get_request_module_name(request))
    request.add_finished_callback(on_request_finished)


def on_request_finished(request):
    if not getattr(request, '_opbeat_client', None):
        return

    route_name = get_route_name(request)
    status_code = get_status_code(request)

    request._opbeat_client.end_transaction(route_name, status_code)
